# Emily's Modifier Dictionary
import re

# define your ender here
uniqueEnders = ["LGTS"]


LONGEST_KEY = 1

# fingerspelling dictionary entries for relevant theories
spelling = {
        "A"     : "a",
        "PW"    : "b",
        "KR"    : "c",
        "TK"    : "d",
        "E"     : "e",
        "TP"    : "f",
        "TKPW"  : "g",
        "H"     : "h",
        "EU"    : "i",
        "AOEU"    : "i", # magnum
        "SKWR"  : "j",
        "SKWRAEU" : "j", # magnum
        "K"     : "k",
        "HR"    : "l",
        "PH"    : "m",
        "TPH"   : "n",
        "O"     : "o",
        "P"     : "p",
        "KW"    : "q",
        "R"     : "r",
        "S"     : "s",
        "T"     : "t",
        "U"     : "u",
        "SR"    : "v",
        "W"     : "w",
        "KP"    : "x",
        "KWH"   : "y",
        "STKPW" : "z",
        "STKPWHR" : "z", # magnum
        }

# same as emily-symbols format, but modified for use on the left hand
symbols = {
        "TR"    : ["tab", "backspace", "delete", "escape"],
        "KPWR"  : ["up", "left", "right", "down"],
        "TKPWR" : ["page_up", "home", "end", "page_down"],
        "TKWR"  : ["AudioPlay", "AudioPrev", "AudioNext", "AudioStop"],
        "TKW"   : ["AudioMute", "AudioLowerVolume", "AudioRaiseVolume", ""],
        ""      : ["", "return", "tab", "space"],

        # typable symbols
        "TK"     : ["exclam", "", "notsign", "exclamdown"],
        "TP"     : ["quotedbl", "", "", ""],
        "TKHR"   : ["numbersign", "", "", ""],
        "KPWH"   : ["dollar", "", "", ""],
        "TKPW"   : ["percent", "", "", ""],
        "TWR"    : ["ampersand", "", "", ""],
        "T"      : ["apostrophe", "", "", ""],
        "TPH"    : ["parenleft", "less", "bracketleft", "braceleft"],
        "KWR"    : ["parenright", "greater", "bracketright", "braceright"],
        "H"      : ["asterisk", "", "", "multiply"],
        "R"      : ["plus", "", "", ""],
        "W"      : ["comma", "", "", ""],
        "PH"     : ["minus", "", "", ""],
        "K"      : ["period", "", "", ""],
        "KP"     : ["slash", "", "", "division"],
        "HR"     : ["colon", "", "", ""],
        "KW"     : ["semicolon", "", "", ""],
        "PWHR"   : ["equal", "", "", ""],
        "TPW"    : ["question", "questiondown", "", ""],
        "TKPWHR" : ["at", "", "", ""],
        "TW"     : ["backslash", "", "", ""],
        "KPR"    : ["asciicircum", "guillemotleft", "guillemotright", ""],
        "WR"     : ["underscore", "", "", "mu"],
        "P"      : ["grave", "", "", ""],
        "PW"     : ["bar", "", "", ""],
        "TPWR"   : ["asciitilde", "", "", ""]
        }

def lookup(chord):

    # extract the chord for easy use
    stroke = chord[0]

    # quick tests to avoid regex if non-relevant stroke is sent
    if len(chord) != 1:
        raise KeyError
    assert len(chord) <= LONGEST_KEY

    # extract relevant parts of the stroke
    firstMatch = re.fullmatch(r'([#STKPWHR]*)([AO]*)([*-]*)([EU]*)([FRPB]*)([LGTSDZ]*)', stroke)

    # error out if there are no matches found
    if firstMatch is None:
        raise KeyError
    # name the relevant extracted parts of the regex
    (key, vowel1, seperator, vowel2, modifiers, ender) = firstMatch.groups()

    # if the user doesn't specify a modifier, then error out as the dictionary has no use otherwise
    if modifiers is None:
        raise KeyError

    # combine the relevant parts of the stroke into a nice name
    pattern = key + vowel1 + vowel2

    if ender not in uniqueEnders:
        raise KeyError

    if pattern == "":
        # just modifiers pressed on their own, used a lot in windows apparently
        character = ""
    else:
        # use * to distinguish symbol input from numerical or character input
        if "*" in seperator:
            # symbol input
            # extract the part of the symbol input
            secondMatch = re.fullmatch(r'([STKPWHR]*)([AO]*)([EU]*)', pattern)
            # into variables
            (pattern, variants, vowel2) = secondMatch.groups()
            # if the pattern is not recognised, error out
            if pattern not in symbols:
                raise KeyError

            # calculate the variant count
            variant = 0
            if 'A' in variants:
                variant = variant + 1
            if 'O' in variants:
                variant = variant + 2

            # get the entry
            entry = symbols[pattern]
            if type(entry) == list:
                extract = entry[variant]
                # error out if the entry isn't applicable
                if extract == "":
                    raise KeyError

                character = extract
            else:
                character = entry
        else:
            # numbers or letters
            # extract relevant parts of the stroke
            secondMatch = re.fullmatch(r'([STKPWHR]*)([AO]*)([-EU]*)', pattern)
            (shape, number, vowel2) = secondMatch.groups()

            # AO is unused in finger spelling, thus used to disginguish numerical input
            if number == "AO" and vowel2 == "":

                # left-hand bottom row counts in binary for numbers 0-9
                count = 0
                if "R" in shape:
                    count = count + 1
                if "W" in shape:
                    count = count + 2
                if "K" in shape:
                    count = count + 4
                if "S" in shape:
                    count = count + 8

                # if TP is being held as well, then user is inputting a Fx key - like alt+F4
                function = False
                if "T" in shape and "P" in shape:
                    function = True

                # add the 'F' if F number
                if function:
                    character = "F" + str(count)
                    if count > 12:
                        raise KeyError
                else:
                    if count > 9:
                        raise KeyError
                    character = str(count)
            else:
                # finger spelling input
                entry = shape + number + vowel2

                # check for entry in dictionary
                if entry not in spelling:
                    raise KeyError
                character = spelling[entry]

                # if there is no entry, pass the error
                if character == "":
                    raise KeyError

    # accumulate list of modifiers to be added to the character
    # may need to reorder?
    modKeys = modifiers
    mods = []

    if len(modKeys) == 0:
        raise KeyError

    if "R" in modKeys:
        mods.append("shift")
    if "B" in modKeys:
        mods.append("alt")
    if "P" in modKeys:
        mods.append("super")
    if "F" in modKeys:
        mods.append("control")

    # apply those modifiers
    combo = character
    for mod in mods:
        combo = mod + "(" + combo + ")"

    # package it up with the syntax
    ret = "{^}{#" + combo + "}{^}"

    # all done! :D
    return ret



LONGEST_KEY = 2

# binary list of modifiers for first stroke, ctrl = 1, alt = 2, shift = 4, super = 8
modifiers = [
    "TKPWHR",  # no modifier keys, press key alone
    "KHR",  # ctrl
    "THRA",  # alt
    "TKHRA",  # ctrl alt
    "SH",  # shift
    "SKHR",  # ctrl shift
    "STHRA",  # alt shift
    "STKHRA",  # ctrl alt shift
    "KPW",  # super
    "KPWHR",  # ctrl super
    "TKPWHRA",  # alt super
    "TKPWHRAO",  # ctrl alt super
    "SKPWHR",  # shift super
    "SKPWHRO",  # ctrl shift super
    "STKPWHRA",  # alt shift super
    "STKPWHRAO"  # ctrl alt shift super
]

# list of keys for second stroke
keys = {
    # letters: fingerspelling
    "A": "a",
    "PW": "b",
    "KR": "c",
    "TK": "d",
    "E": "e",
    "SK": "e",  # left hand stroke for E
    "TP": "f",
    "TKPW": "g",
    "H": "h",
    "EU": "i",
    "SKW": "i",  # left hand stroke for I
    "SKWR": "j",
    "K": "k",
    "HR": "l",
    "PH": "m",
    "TPH": "n",
    "O": "o",
    "P": "p",
    "KW": "q",
    "R": "r",
    "S": "s",
    "T": "t",
    "U": "u",
    "WR": "u",  # left hand stroke for U
    "SR": "v",
    "W": "w",
    "KP": "x",
    "KWH": "y",
    "STK": "z",  # simpler stroke for Z
    "STKPW": "z",
    # numbers: AO + binary
    "AO": "0",
    "RAO": "1",
    "WAO": "2",
    "WRAO": "3",
    "KAO": "4",
    "KRAO": "5",
    "KWAO": "6",
    "KWRAO": "7",
    "SAO": "8",
    "SRAO": "9",
    # function keys: F + binary
    "TPRAO": "f1",
    "TPWAO": "f2",
    "TPWRAO": "f3",
    "TKPAO": "f4",
    "TKPRAO": "f5",
    "TKPWAO": "f6",
    "TKPWRAO": "f7",
    "STPAO": "f8",
    "STPRAO": "f9",
    "STPWAO": "f10",
    "STPWRAO": "f11",
    "STKPAO": "f12",
    # numpad keys: asterisk + non-numpad key
    "AO*": "kp_0",
    "RAO*": "kp_1",
    "WAO*": "kp_2",
    "WRAO*": "kp_3",
    "KAO*": "kp_4",
    "KRAO*": "kp_5",
    "KWAO*": "kp_6",
    "KWRAO*": "kp_7",
    "SAO*": "kp_8",
    "SRAO*": "kp_9",
    "PR*": "kp_decimal",  # PeRiod + asterisk
    "PHR*": "kp_add",  # PLus + asterisk
    "PHR": "kp_add",  # PHR isn't used by anything else
    "TKH*": "kp_subtract",  # DasH + asterisk
    "STR*": "kp_multiply",  # STaR + asterisk
    "STR": "kp_multiply",  # STR isn't used by anything else
    "SHR*": "kp_divide",  # SLash + asterisk
    "TPH*": "num_lock",  # Num lock + asterisk
    # symbol keys
    "PR": "period",  # PeRiod
    "KPH": "comma",  # CoMMa
    "SHR": "slash",  # SLash
    "SKHR": "semicolon",  # SemiCoLon
    "KWO": "apostrophe",  # QUOte
    "PWHR": "bracketleft",  # Bracket Left
    "PWR": "bracketright",  # Bracket Right
    "SPWHR": "backslash",  # BackSLash
    "TKH": "minus",  # DasH
    "KWA": "equal",  # eQUAl
    "TKPWR": "grave",  # GRave
    # arrows: PKWR + A (Arrow)
    "PA": "up",
    "WA": "down",
    "KA": "left",
    "RA": "right",
    # navigation keys: arrows + asterisk
    "PA*": "prior",  # page up
    "WA*": "next",  # page down
    "KA*": "home",
    "RA*": "end",
    # misc keys
    "KHR*": "caps_lock",  # Caps Lock
    "SKHR*": "scroll_lock",  # SCroll Lock + asterisk (SKHR taken by semicolon)
    "SP": "space",  # SPace
    "SPWR": "return",  # ENTeR
    "TPW": "tab",  # TaB
    "SPW": "backspace",  # BackSpace
    "TKHR": "delete",  # DeLete
    "SKA": "escape",  # eSCApe
    "STPH": "insert",  # iNSert
    "SKPR": "print",  # SCreen PRint
    "PH*": "menu",  # Menu + asterisk
    # "KPO": "multi_key",  # ComPOse, enable only if you have a compose key
    "TPHO": "" #modifier keys alone
}


def lookup(outline):
    assert len(outline) <= LONGEST_KEY
    str1 = outline[0]
    # KeyError if first stroke is not one of the modifiers
    if str1 not in modifiers:
        raise KeyError

    mods = modifiers.index(str1)  # get index of modifier keys
    # length should be 2
    assert len(outline) == 2
    str2 = outline[1]
    # do nothing if second stroke is SKPH
    if str2 == "SKPH":
        return "{#}"
    if str2 not in keys:
        raise KeyError
    assert str2 in keys
    # get key to press from second stroke
    key = keys.get(str2)

    # add modifiers
    # 8 = super
    if mods & 8 == 8:
        key = "super_l(" + key + ")"
    # 4 = shift
    if mods & 4 == 4:
        key = "shift_l(" + key + ")"
    # 2 = alt
    if mods & 2 == 2:
        key = "alt_l(" + key + ")"
    # 1 = control
    if mods & 1 == 1:
        key = "control_l(" + key + ")"

    # add syntax
    key = "{#" + key + "}"

    # remove unnecessary parentheses
    key = key.replace("()", "")

    # done!
    return key



# Jeff's phrasing dictionary for Plover.

# See README.md for instructions on how the system works.

import re

LONGEST_KEY = 1

PARTS_MATCHER = re.compile(
    r'(S?T?K?P?W?H?R?)(A?O?)-?(\*?)(E?U?)(F?)(R?P?B?L?G?T?S?D?Z?)'
)

TO_BE = {
    "present": {
        None: " are",
        "root": " be",
        "1ps": " am",
        "3ps": " is",
        "present-participle": " being",
        "past-participle": " been",
    },
    "past": {
        None: " were",
        "root": " be",
        "1ps": " was",
        "3ps": " was",
        "present-participle": " being",
        "past-participle": " been",
    },
}

TO_HAVE = {
    "present": {
        None: " have",
        "3ps": " has",
        "present-participle": " having",
        "past-participle": " had",
    },
    "past": {
        "root": " have",
        None: " had",
        "present-participle": " having",
        "past-participle": " had",
    }
}

THERE_SUFFIXES = {
    "": True,
    "D": True,                                                # Past tense
    "B": True, "BT": True, "BD": True, "BTD": True,           # Be (a)
    "BG": True, "BGD": True,                                  # Come
    "G": True, "GD": True,                                    # Go
    "PZ": True, "PDZ": True,                                  # Happen
    "T": True, "TD": True, "TS": True, "TSDZ": True,          # Have (to)
    "LZ": True, "LZD": True,                                  # Live
    "PL": True, "PLT": True, "PLD": True, "PLTD": True,       # May (have)
    "PBLGS": True, "PBLGTS": True,                            # Must (have)
    "PBLGSZ": True, "PBLGTSDZ": True,                         # Just
    "RPG": True, "RPGD": True, "RPGT": True, "RPGTD": True,   # Need (to)
    "RLG": True, "RLGD": True,                                # Really
    "PLS": True, "PLSZ": True, "PLTS": True, "PLTSDZ": True,  # Seem (to)
    "Z": True, "DZ": True, "TZ": True, "TDZ": True,           # Use (to)
}

NON_PHRASE_STROKES = {
    "STHR": True,        # "is there"
    "STHRET": True,      # "stiletto"
    "STHREUPLT": True,   # "stimulate"
    "STPHREFPLT": True,  # "investment in"

    "SKPUR": True,       # "and you're" -- rather than "and you run"
    "SKPUL": True,       # "and you'll" -- rather than "and you look"
    "SKPEUT": True,      # "and it"     -- rather than "and I have"

    "SKP*": True,        # {&&}
}

STARTERS = {
    # Map of stroke -> word, verb-form, valid enders
    #  * 'b' for blank
    #  * '1p', '2p', '3p' for 1st, 2nd, 3rd person
    #  * 'p/s' for plural/singular
    "SWR": ("I", "1ps", None),
    "KPWR": ("you", "2p", None),
    "KWHR": ("he", "3ps", None),
    "SKWHR": ("she", "3ps", None),
    "KPWH": ("it", "3ps", None),
    "TWR": ("we", "1pp", None),
    "TWH": ("they", "3pp", None),
    "STKH": ("this", "3ps", None),
    "STWH": ("that", "3ps", None),
    "STHR": ("there", "3ps", THERE_SUFFIXES),
    "STPHR": ("there", "3pp", THERE_SUFFIXES),
    "STKPWHR": ("", "b3ps", None),
    "STWR": ("", "b3pp", None),
}

SIMPLE_STARTERS = {
    "STHA": (" that", None),
    "STPA": (" if", None),
    "SWH": (" when", None),
    "SWHA": (" what", None),
    "SWHR": (" where", None),
    "SWHO": (" who", None),
    "SWHAO": (" why", None),

    "SPWH": (" but", None),
    "STPR": (" for", None),

    # Remove the entry below if you don't want "and" phrases.
    "SKP": (" and", None),
}

SIMPLE_PRONOUNS = {
    "E": ("he", "3ps", None),
    "*E": ("she", "3ps", None),
    "U": ("you", "2p", None),
    "*U": ("they", "3pp", None),
    "EU": ("I", "1ps", None),
    "*EU": ("we", "1pp", None),
    "*": ("it", "3ps", None),
}

SIMPLE_STRUCTURES = {
    "": ("* !", True, None),
    "F": ({tense: {form: "* !" + TO_HAVE[tense][form] for form in TO_HAVE[tense]} for tense in TO_HAVE}, True, "past-participle"),
}

MIDDLES = {
    # Map of stroke -> (map[tense][verb-form](string, verb-form)
    "": {"present": {None: (" do", "root"), "3ps": (" does", "root")}, "past": (" did", "root")},
    "*": {"present": {None: (" don't", "root"), "3ps": (" doesn't", "root")}, "past": (" didn't", "root")},
    "A": {"present": (" can", "root"), "past": (" could", "root")},
    "A*": {"present": (" can't", "root"), "past": (" couldn't", "root")},
    "O": {"present": (" shall", "root"), "past": (" should", "root")},
    "O*": {"present": (" shall not", "root"), "past": (" shouldn't", "root")},
    "AO": {"present": (" will", "root"), "past": (" would", "root")},
    "AO*": {"present": (" won't", "root"), "past": (" wouldn't", "root")},
}

# `!` represents the starter
# `*` represents the middle
#
# (format, use-middle-verb-form, updated-verb-form)
STRUCTURE_EXCEPTIONS = {
    "": ("!", False, None),

    # To make reverse look-ups work correctly, longer results must be listed first

    "*E": ({"present": {None: "! aren't", "1ps": "! am not", "3ps": "! isn't"}, "past": {None: "! weren't", "1ps": "! wasn't", "3ps": "! wasn't"}}, False, "present-participle"),
    "E": ({"present": {None: "! are", "1ps": "! am", "3ps": "! is"}, "past": {None: "! were", "1ps": "! was", "3ps": "! was"}}, False, "present-participle"),
    "*F": ({"present": {None: "! haven't", "3ps": "! hasn't"}, "past": "! hadn't"}, False, "past-participle"),
    "F": ({"present": {None: "! have", "3ps": "! has"}, "past": "! had"}, False, "past-participle"),
    "*EF": ({"present": {None: "! haven't been", "3ps": "! hasn't been"}, "past": "! hadn't been"}, False, "present-participle"),
    "EF": ({"present": {None: "! have been", "3ps": "! has been"}, "past": "! had been"}, False, "present-participle"),

    # The following alternate definitions cause phrases to use contracted forms, e.g.:
    #
    # * `I am going to` -> `I'm going to`
    # * `I have gone to` -> `I've gone to`
    #
    # To use contracted forms, uncomment out the next 6 definitions.
    #
    # "*E": ({"present": {None: "! are not", "1ps": "!'m not", "2p": "!'re not", "3ps": "! isn't", "1pp": "!'re not", "3pp": "!'re not", "b3pp": "! are not"}, "past": {None: "! weren't", "1ps": "! wasn't", "3ps": "! wasn't"}}, False, "present-participle"),
    # "E": ({"present": {None: "! are", "1ps": "!'m", "2p": "!'re", "3ps": "!'s", "b3ps": "! is", "1pp": "!'re", "3pp": "!'re", "b3pp": "! are"}, "past": {None: "! were", "1ps": "! was", "3ps": "! was"}}, False, "present-participle"),
    # "*F": ({"present": {None: "! haven't", "3ps": "! hasn't"}, "past": "! hadn't"}, False, "past-participle"),
    # "F": ({"present": {None: "! have", "1ps": "!'ve", "2p": "!'ve", "3ps": "!'s", "b3ps": "! has", "1pp": "!'ve", "3pp": "!'ve", "b3pp": "! have"}, "past": {None: "! had", "1ps": "!'d", "2p": "!'d", "3ps": "!'d", "1pp": "!'d", "3pp": "!'d", "b3pp": "! had"}}, False, "past-participle"),
    # "*EF": ({"present": {None: "! haven't been", "3ps": "! hasn't been"}, "past": "! hadn't been"}, False, "present-participle"),
    # "EF": ({"present": {None: "! have been", "1ps": "!'ve been", "2p": "!'ve been", "3ps": "!'s been", "b3ps": "! has been", "1pp": "!'ve been", "3pp": "!'ve been", "b3pp": "! have been"}, "past": {None: "! had been", "1ps": "!'d been", "2p": "!'d been", "3ps": "!'d been", "b3ps": "! had been", "1pp": "!'d been", "3pp": "!'d been", "b3pp": "! had been"}}, False, "present-participle"),

    # This is an old system that used 'U' for inverting have/be.
    # It's been commented out as the never/still/just
    # "*EU": ({"present": {None: " aren't !", "1ps": " am ! not", "3ps": " isn't !"}, "past": {None: " weren't !", "1ps": " wasn't !", "3ps": " wasn't !"}}, False, "present-participle"),
    # "EU": ({"present": {None: " are !", "1ps": " am !", "3ps": " is !"}, "past": {None: " were !", "1ps": " was !", "3ps": " was !"}}, False, "present-participle"),
    # "*UF": ({"present": {None: " haven't !", "3ps": " hasn't !"}, "past": " hadn't !"}, False, "past-participle"),
    # "UF": ({"present": {None: " have !", "3ps": " has !"}, "past": " had !"}, False, "past-participle"),
    # "*EUF": ({"present": {None: " haven't ! been", "3ps": " hasn't ! been"}, "past": " hadn't ! been"}, False, "present-participle"),
    # "EUF": ({"present": {None: " have ! been", "3ps": " has ! been"}, "past": " had ! been"}, False, "present-participle"),

    "EU": ("! still", False, None),
    "EUF": ("! never", False, None),
    "UF": ("! just", False, None),

    # Special cases for empty starters.
    # - infinitive forms.
    "STWRU": ("to", False, "root"),
    "STWR*U": ("not to", False, "root"),
    "STKPWHRU": ("to", False, "root"),
    "STKPWHR*U": ("not to", False, "root"),
}

ALWAYS = {None: "* !", "b3ps-root": "* always", "b3pp-root": "* always"}

STRUCTURES = {
    "": ("!*", True, None),
    "*": ("!*", True, None),

    "*E": ({tense: {form: "!*" + TO_BE[tense][form] for form in TO_BE[tense]} for tense in TO_BE}, True, "present-participle"),
    "E": ({tense: {form: "!*" + TO_BE[tense][form] for form in TO_BE[tense]} for tense in TO_BE}, True, "present-participle"),
    "*EF": ({tense: {form: "!*" + TO_HAVE[tense][form] + " been" for form in TO_HAVE[tense]} for tense in TO_HAVE}, True, "present-participle"),
    "EF": ({tense: {form: "!*" + TO_HAVE[tense][form] + " been" for form in TO_HAVE[tense]} for tense in TO_HAVE}, True, "present-participle"),
    "*F": ({tense: {form: "!*" + TO_HAVE[tense][form] for form in TO_HAVE[tense]} for tense in TO_HAVE}, True, "past-participle"),
    "F": ({tense: {form: "!*" + TO_HAVE[tense][form] for form in TO_HAVE[tense]} for tense in TO_HAVE}, True, "past-participle"),

    "*EU": ("! still*", True, None),
    "EU": ("!* still", True, None),
    "*EUF": ("!* even", True, None),
    "EUF": ("!* never", True, None),

    "*U": ({"present": ALWAYS, "past": ALWAYS}, True, None),
    "U": ({"present": ALWAYS, "past": ALWAYS}, True, None),
    "*UF": ("! just*", True, None),
    "UF": ("!* just", True, None),
} 

ENDERS = {
    "": ("present", ""),
    "D": ("past", ""),

    # RB: To ask
    "RB": ("present", {None: " ask", "3ps": " asks", "present-participle": " asking", "past-participle": " asked"}),
    "RBD": ("past", {None: " asked", "root": " ask", "present-participle": " asking", "past-participle": " asked"}),

    # B: To be (a)
    "B": ("present", TO_BE["present"]),
    "BT": ("present", {key: TO_BE["present"][key] + " a" for key in TO_BE["present"]}),
    "BD": ("past", TO_BE["past"]),
    "BTD": ("past", {key: TO_BE["past"][key] + " a" for key in TO_BE["past"]}),

    # RPBG: To become (a)
    "RPBG": ("present", {None: " become", "3ps": " becomes", "present-participle": " becoming", "past-participle": " become"}),
    "RPBGT": ("present", {None: " become a", "3ps": " becomes a", "present-participle": " becoming a", "past-participle": " become a"}),
    "RPBGD": ("past", {None: " became", "root": " become", "present-participle": " becoming", "past-participle": " become"}),
    "RPBGTD": ("past", {None: " became a", "root": " become a", "present-participle": " becoming a", "past-participle": " become a"}),

    # BL - To believe (that)
    "BL": ("present", {None: " believe", "3ps": " believes", "present-participle": " believing", "past-participle": " believed"}),
    "BLT": ("present", {None: " believe that", "3ps": " believes that", "present-participle": " believing that", "past-participle": " believed that"}),
    "BLD": ("past", {None: " believed", "root": " believe", "present-participle": " believing", "past-participle": " believed"}),
    "BLTD": ("past", {None: " believed that", "root": " believe that", "present-participle": " believing that", "past-participle": " believed that"}),

    # RBLG - To call
    "RBLG": ("present", {None: " call", "3ps": " calls", "present-participle": " calling", "past-participle": " called"}),
    "RBLGD": ("past", {None: " called", "root": " call", "present-participle": " calling", "past-participle": " called"}),

    # BGS - Can - Auxiliary verb
    # These do not combine naturally with middle/structures.
    "BGS": ("present", " can"),
    "BGSZ": ("past", " could"),

    # RZ - To care
    "RZ": ("present", {None: " care", "3ps": " cares", "present-participle": " caring", "past-participle": " cared"}),
    "RDZ": ("past", {None: " cared", "root": " care", "present-participle": " caring", "past-participle": " cared"}),

    # PBGZ - To change
    "PBGZ": ("present", {None: " change", "3ps": " changes", "present-participle": " changing", "past-participle": " changed"}),
    "PBGDZ": ("past", {None: " changed", "root": " change", "present-participle": " changing", "past-participle": " changed"}),

    # BG - To come (to)
    "BG": ("present", {None: " come", "3ps": " comes", "present-participle": " coming", "past-participle": " come"}),
    "BGT": ("present", {None: " come to", "3ps": " comes to", "present-participle": " coming to", "past-participle": " come to"}),
    "BGD": ("past", {None: " came", "root": " come", "present-participle": " coming", "past-participle": " come"}),
    "BGTD": ("past", {None: " came to", "root": " come to", "present-participle": " coming to", "past-participle": " come to"}),

    # RBGZ - To consider
    "RBGZ": ("present", {None: " consider", "3ps": " considers", "present-participle": " considering", "past-participle": " considered"}),
    "RBGDZ": ("past", {None: " considered", "root": " consider", "present-participle": " considering", "past-participle": " considered"}),

    # RP - To do (it)
    "RP": ("present", {None: " do", "3ps": " does", "present-participle": " doing", "past-participle": " done"}),
    "RPT": ("present", {None: " do it", "3ps": " does it", "present-participle": " doing it", "past-participle": " done it"}),
    "RPD": ("past", {None: " did", "root": " do", "present-participle": " doing", "past-participle": " done"}),
    "RPTD": ("past", {None: " did it", "root": " do it", "present-participle": " doing it", "past-participle": " done it"}),

    # PGS: To expect (that)
    "PGS": ("present", {None: " expect", "3ps": " expects", "present-participle": " expecting", "past-participle": " expected"}),
    "PGTS": ("present", {None: " expect that", "3ps": " expects that", "present-participle": " expecting that", "past-participle": " expected that"}),
    "PGSZ": ("past", {None: " expected", "root": " expect", "present-participle": " expecting", "past-participle": " expected"}),
    "PGTSDZ": ("past", {None: " expected that", "root": " expect that", "present-participle": " expecting that", "past-participle": " expected that"}),

    # LT - To feel (like)
    "LT": ("present", {None: " feel", "3ps": " feels", "present-participle": " feeling", "past-participle": " felt"}),
    "LTS": ("present", {None: " feel like", "3ps": " feels like", "present-participle": " feeling like", "past-participle": " felt like"}),
    "LTD": ("past", {None: " felt", "root": " feel", "present-participle": " feeling", "past-participle": " felt"}),
    "LTSDZ": ("past", {None: " felt like", "root": " feel like", "present-participle": " feeling like", "past-participle": " felt like"}),

    # PBLG - To find (that)
    "PBLG": ("present", {None: " find", "3ps": " finds", "present-participle": " finding", "past-participle": " found"}),
    "PBLGT": ("present", {None: " find that", "3ps": " finds that", "present-participle": " finding that", "past-participle": " found that"}),
    "PBLGD": ("past", {None: " found", "root": " find", "present-participle": " finding", "past-participle": " found"}),
    "PBLGTD": ("past", {None: " found that", "root": " find that", "present-participle": " finding that", "past-participle": " found that"}),

    # RG: To forget (to)
    "RG": ("present", {None: " forget", "3ps": " forgets", "present-participle": " forgetting", "past-participle": " forgotten"}),
    "RGT": ("present", {None: " forget to", "3ps": " forgets to", "present-participle": " forgetting to", "past-participle": " forgotten to"}),
    "RGD": ("past", {None: " forgot", "root": " forget", "present-participle": " forgetting", "past-participle": " forgotten"}),
    "RGTD": ("past", {None: " forgot to", "root": " forget to", "present-participle": " forgetting to", "past-participle": " forgotten to"}),

    # GS: To get (to)
    "GS": ("present", {None: " get", "3ps": " gets", "present-participle": " getting", "past-participle": " got"}),
    "GTS": ("present", {None: " get to", "3ps": " gets to", "present-participle": " getting to", "past-participle": " got to"}),
    "GSZ": ("past", {None: " got", "root": " get", "present-participle": " getting", "past-participle": " got"}),
    "GTSDZ": ("past", {None: " got to", "root": " get to", "present-participle": " getting to", "past-participle": " got to"}),

    # GZ: To give
    "GZ": ("present", {None: " give", "3ps": " gives", "present-participle": " giving", "past-participle": " given"}),
    "GDZ": ("past", {None: " gave", "root": " give", "present-participle": " giving", "past-participle": " given"}),

    # G: To go (to)
    "G": ("present", {None: " go", "3ps": " goes", "present-participle": " going", "past-participle": " gone"}),
    "GT": ("present", {None: " go to", "3ps": " goes to", "present-participle": " going to", "past-participle": " gone to"}),
    "GD": ("past", {None: " went", "root": " go", "present-participle": " going", "past-participle": " gone"}),
    "GTD": ("past", {None: " went to", "root": " go to", "present-participle": " going to", "past-participle": " gone to"}),

    # T - To have (to)
    "T": ("present", {None: " have", "3ps": " has", "present-participle": " having", "past-participle": " had"}),
    "TS": ("present", {None: " have to", "3ps": " has to", "present-participle": " having to", "past-participle": " had to"}),
    "TD": ("past", {None: " had", "root": " have", "present-participle": " having", "past-participle": " had"}),
    "TSDZ": ("past", {None: " had to", "root": " have to", "present-participle": " having to", "past-participle": " had to"}),

    # PZ - To happen
    "PZ": ("present", {None: " happen", "3ps": " happens", "present-participle": " happening", "past-participle": " happened"}),
    "PDZ": ("past", {None: " happened", "root": " happen", "present-participle": " happening", "past-participle": " happened"}),

    # PG - To hear (that)
    "PG": ("present", {None: " hear", "3ps": " hears", "present-participle": " hearing", "past-participle": " heard"}),
    "PGT": ("present", {None: " hear that", "3ps": " hears that", "present-participle": " hearing that", "past-participle": " heard that"}),
    "PGD": ("past", {None: " heard", "root": " hear", "present-participle": " hearing", "past-participle": " heard"}),
    "PGTD": ("past", {None: " heard that", "root": " hear that", "present-participle": " hearing that", "past-participle": " heard that"}),

    # RPS - To hope (to)
    "RPS": ("present", {None: " hope", "3ps": " hopes", "present-participle": " hoping", "past-participle": " hoped"}),
    "RPTS": ("present", {None: " hope to", "3ps": " hopes to", "present-participle": " hoping to", "past-participle": " hoped to"}),
    "RPSZ": ("past", {None: " hoped", "root": " hope", "present-participle": " hoping", "past-participle": " hoped"}),
    "RPTSDZ": ("past", {None: " hoped to", "root": " hope to", "present-participle": " hoping to", "past-participle": " hoped to"}),

    # PLG - To imagine (that)
    "PLG": ("present", {None: " imagine", "3ps": " imagines", "present-participle": " imagining", "past-participle": " imagined"}),
    "PLGT": ("present", {None: " imagine that", "3ps": " imagines that", "present-participle": " imagining that", "past-participle": " imagined that"}),
    "PLGD": ("past", {None: " imagined", "root": " imagine", "present-participle": " imagining", "past-participle": " imagined"}),
    "PLGTD": ("past", {None: " imagined that", "root": " imagine that", "present-participle": " imagining that", "past-participle": " imagined that"}),

    # PBLGSZ - Just
    "PBLGSZ": ("present", " just"),
    "PBLGTSDZ": ("past", " just"),

    # PBGS - To keep
    "PBGS": ("present", {None: " keep", "3ps": " keeps", "present-participle": " keeping", "past-participle": " kept"}),
    "PBGSZ": ("past", {None: " kept", "root": " keep", "present-participle": " keeping", "past-participle": " kept"}),

    # PB: To know (that)
    "PB": ("present", {None: " know", "3ps": " knows", "present-participle": " knowing", "past-participle": " known"}),
    "PBT": ("present", {None: " know that", "3ps": " knows that", "present-participle": " knowing that", "past-participle": " known that"}),
    "PBD": ("past", {None: " knew", "root": " know", "present-participle": " knowing", "past-participle": " known"}),
    "PBTD": ("past", {None: " knew that", "root": " know that", "present-participle": " knowing that", "past-participle": " known that"}),

    # RPBS - To learn (to)
    "RPBS": ("present", {None: " learn", "3ps": " learns", "present-participle": " learning", "past-participle": " learned"}),
    "RPBTS": ("present", {None: " learn to", "3ps": " learns to", "present-participle": " learning to", "past-participle": " learned to"}),
    "RPBSZ": ("past", {None: " learned", "root": " learn", "present-participle": " learning", "past-participle": " learned"}),
    "RPBTSDZ": ("past", {None: " learned to", "root": " learn to", "present-participle": " learning to", "past-participle": " learned to"}),

    # LGZ - To leave
    "LGZ": ("present", {None: " leave", "3ps": " leaves", "present-participle": " leaving", "past-participle": " left"}),
    "LGDZ": ("past", {None: " left", "root": " leave", "present-participle": " leaving", "past-participle": " left"}),

    # LS - To let
    "LS": ("present", {None: " let", "3ps": " lets", "present-participle": " letting", "past-participle": " let"}),
    "LSZ": ("past", {None: " let", "present-participle": " letting", "past-participle": " let"}),

    # BLG - To like (to)
    "BLG": ("present", {None: " like", "3ps": " likes", "present-participle": " liking", "past-participle": " liked"}),
    "BLGT": ("present", {None: " like to", "3ps": " likes to", "present-participle": " liking to", "past-participle": " liked to"}),
    "BLGD": ("past", {None: " liked", "root": " like", "present-participle": " liking", "past-participle": " liked"}),
    "BLGTD": ("past", {None: " liked to", "root": " like to", "present-participle": " liking to", "past-participle": " liked to"}),

    # LZ - To live
    "LZ": ("present", {None: " live", "3ps": " lives", "present-participle": " living", "past-participle": " lived"}),
    "LDZ": ("past", {None: " lived", "root": " live", "present-participle": " living", "past-participle": " lived"}),

    # L - To look
    "L": ("present", {None: " look", "3ps": " looks", "present-participle": " looking", "past-participle": " looked"}),
    "LD": ("past", {None: " looked", "root": " look", "present-participle": " looking", "past-participle": " looked"}),

    # LG - To love (to)
    "LG": ("present", {None: " love", "3ps": " loves", "present-participle": " loving", "past-participle": " loved"}),
    "LGT": ("present", {None: " love to", "3ps": " loves to", "present-participle": " loving to", "past-participle": " loved to"}),
    "LGD": ("past", {None: " loved", "root": " love", "present-participle": " loving", "past-participle": " loved"}),
    "LGTD": ("past", {None: " loved to", "root": " love to", "present-participle": " loving to", "past-participle": " loved to"}),

    # RPBL - To make (a)
    "RPBL": ("present", {None: " make", "3ps": " makes", "present-participle": " making", "past-participle": " made"}),
    "RPBLT": ("present", {None: " make a", "3ps": " makes a", "present-participle": " making a", "past-participle": " made a"}),
    "RPBLD": ("past", {None: " made", "root": " make", "present-participle": " making", "past-participle": " made"}),
    "RPBLTD": ("past", {None: " made a", "root": " make a", "present-participle": " making a", "past-participle": " made a"}),

    # PL - Auxiliary verb may (be)
    # These do not combine naturally with middle/structures.
    "PL": ("present", " may"),
    "PLT": ("present", " may be"),
    "PLD": ("past", " might"),
    "PLTD": ("past", " might be"),

    # PBL - To mean (to)
    "PBL": ("present", {None: " mean", "3ps": " means", "present-participle": " meaning", "past-participle": " meant"}),
    "PBLT": ("present", {None: " mean to", "3ps": " means to", "present-participle": " meaning to", "past-participle": " meant to"}),
    "PBLD": ("past", {None: " meant", "root": " mean", "present-participle": " meaning", "past-participle": " meant"}),
    "PBLTD": ("past", {None: " meant to", "root": " mean to", "present-participle": " meaning to", "past-participle": " meant to"}),

    # PBLS - To mind
    "PBLS": ("present", {None: " mind", "3ps": " minds", "present-participle": " minding", "past-participle": " minded"}),
    "PBLSZ": ("past", {None: " minded", "root": " mind", "present-participle": " minding", "past-participle": " minded"}),

    # PLZ - To move
    "PLZ": ("present", {None: " move", "3ps": " moves", "present-participle": " moving", "past-participle": " moved"}),
    "PLDZ": ("past", {None: " moved", "root": " move", "present-participle": " moving", "past-participle": " moved"}),

    # PBLGS - Auxiliary verb must (be)
    # These do not combine naturally with middle/structures.
    "PBLGS": ("present", " must"),
    "PBLGTS": ("present", " must be"),

    # RPG: To need (to)
    "RPG": ("present", {None: " need", "3ps": " needs", "present-participle": " needing", "past-participle": " needed"}),
    "RPGT": ("present", {None: " need to", "3ps": " needs to", "present-participle": " needing to", "past-participle": " needed to"}),
    "RPGD": ("past", {None: " needed", "root": " need", "present-participle": " needing", "past-participle": " needed"}),
    "RPGTD": ("past", {None: " needed to", "root": " need to", "present-participle": " needing to", "past-participle": " needed to"}),

    # PS: To put (it)
    "PS": ("present", {None: " put", "3ps": " puts", "present-participle": " putting", "past-participle": " put"}),
    "PTS": ("present", {None: " put it", "3ps": " puts it", "present-participle": " putting it", "past-participle": " put it"}),
    "PSZ": ("past", {None: " put", "root": " put", "present-participle": " putting", "past-participle": " put"}),
    "PTSDZ": ("past", {None: " put it", "root": " put it", "present-participle": " putting it", "past-participle": " put it"}),

    # RS - To read
    "RS": ("present", {None: " read", "3ps": " reads", "present-participle": " reading", "past-participle": " read"}),
    "RSZ": ("past", {None: " read", "root": " read", "present-participle": " reading", "past-participle": " read"}),

    # RLG - really
    "RLG": ("present", " really"),
    "RLGD": ("past", " really"),

    # RL - To recall
    "RL": ("present", {None: " recall", "3ps": " recalls", "present-participle": " recalling", "past-participle": " recalled"}),
    "RLD": ("past", {None: " recalled", "root": " recall", "present-participle": " recalling", "past-participle": " recalled"}),

    # RLS - To realize (that)
    "RLS": ("present", {None: " realize", "3ps": " realizes", "present-participle": " realizing", "past-participle": " realized"}),
    "RLTS": ("present", {None: " realize that", "3ps": " realizes that", "present-participle": " realizing that", "past-participle": " realized that"}),
    "RLSZ": ("past", {None: " realized", "root": " realize", "present-participle": " realizing", "past-participle": " realized"}),
    "RLTSDZ": ("past", {None: " realized that", "root": " realize that", "present-participle": " realizing that", "past-participle": " realized that"}),

    # RPL - To remember (that)
    "RPL": ("present", {None: " remember", "3ps": " remembers", "present-participle": " remembering", "past-participle": " remembered"}),
    "RPLT": ("present", {None: " remember that", "3ps": " remembers that", "present-participle": " remembering that", "past-participle": " remembered that"}),
    "RPLD": ("past", {None: " remembered", "root": " remember", "present-participle": " remembering", "past-participle": " remembered"}),
    "RPLTD": ("past", {None: " remembered that", "root": " remember that", "present-participle": " remembering that", "past-participle": " remembered that"}),

    # RPLS: To remain
    "RPLS": ("present", {None: " remain", "3ps": " remains", "present-participle": " remaining", "past-participle": " remained"}),
    "RPLSZ": ("past", {None: " remained", "root": " remain", "present-participle": " remaining", "past-participle": " remained"}),

    # R - To run
    "R": ("present", {None: " run", "3ps": " runs", "present-participle": " running", "past-participle": " run"}),
    "RD": ("past", {None: " ran", "root": " run", "present-participle": " running", "past-participle": " run"}),

    # BS - To say (that)
    "BS": ("present", {None: " say", "3ps": " says", "present-participle": " saying", "past-participle": " said"}),
    "BTS": ("present", {None: " say that", "3ps": " says that", "present-participle": " saying that", "past-participle": " said that"}),
    "BSZ": ("past", {None: " said", "root": " say", "present-participle": " saying", "past-participle": " said"}),
    "BTSDZ": ("past", {None: " said that", "root": " say that", "present-participle": " saying that", "past-participle": " said that"}),

    # S - To see
    "S": ("present", {None: " see", "3ps": " sees", "present-participle": " seeing", "past-participle": " seen"}),
    "SZ": ("past", {None: " saw", "root": " see", "present-participle": " seeing", "past-participle": " seen"}),

    # BLS - To set
    "BLS": ("present", {None: " set", "3ps": " sets", "present-participle": " setting", "past-participle": " set"}),
    "BLSZ": ("past", {None: " set", "root": " set", "present-participle": " setting", "past-participle": " set"}),

    # PLS - To seem (to)
    "PLS": ("present", {None: " seem", "3ps": " seems", "present-participle": " seeming", "past-participle": " seemed"}),
    "PLTS": ("present", {None: " seem to", "3ps": " seems to", "present-participle": " seeming to", "past-participle": " seemed to"}),
    "PLSZ": ("past", {None: " seemed", "root": " seem", "present-participle": " seeming", "past-participle": " seemed"}),
    "PLTSDZ": ("past", {None: " seemed to", "root": " seem to", "present-participle": " seeming to", "past-participle": " seemed to"}),

    # RBL - shall - Auxiliary verb
    # These do not combine naturally with middle/structures.
    "RBL": ("present", " shall"),
    "RBLD": ("past", " should"),

    # RBZ - To show
    "RBZ": ("present", {None: " show", "3ps": " shows", "present-participle": " showing", "past-participle": " shown"}),
    "RBDZ": ("past", {None: " showed", "root": " show", "present-participle": " showing", "past-participle": " shown"}),

    # RBT - To take
    "RBT": ("present", {None: " take", "3ps": " takes", "present-participle": " taking", "past-participle": " taken"}),
    "RBTD": ("past", {None: " took", "root": " take", "present-participle": " taking", "past-participle": " taken"}),

    # BLGT - To talk -- conflict with like to.
    # "BLGT": ("present", {None: " talk", "3ps": " talks", "present-participle": " talking", "past-participle": " talked"}),
    # "BLGTD": ("past", {None: " talked", "root": " talk", "present-participle": " talking", "past-participle": " talked"}),

    # RLT - To tell
    "RLT": ("present", {None: " tell", "3ps": " tells", "present-participle": " telling", "past-participle": " told"}),
    "RLTD": ("past", {None: " told", "root": " tell", "present-participle": " telling", "past-participle": " told"}),

    # PBG - To think (that)
    "PBG": ("present", {None: " think", "3ps": " thinks", "present-participle": " thinking", "past-participle": " thought"}),
    "PBGT": ("present", {None: " think that", "3ps": " thinks that", "present-participle": " thinking that", "past-participle": " thought that"}),
    "PBGD": ("past", {None: " thought", "root": " think", "present-participle": " thinking", "past-participle": " thought"}),
    "PBGTD": ("past", {None: " thought that", "root": " think that", "present-participle": " thinking that", "past-participle": " thought that"}),

    # RT - To try (to)
    "RT": ("present", {None: " try", "3ps": " tries", "present-participle": " trying", "past-participle": " tried"}),
    "RTS": ("present", {None: " try to", "3ps": " tries to", "present-participle": " trying to", "past-participle": " tried to"}),
    "RTD": ("past", {None: " tried", "root": " try", "present-participle": " trying", "past-participle": " tried"}),
    "RTSDZ": ("past", {None: " tried to", "root": " try to", "present-participle": " trying to", "past-participle": " tried to"}),

    # RPB - To understand (the)
    "RPB": ("present", {None: " understand", "3ps": " understands", "present-participle": " understanding", "past-participle": " understood"}),
    "RPBT": ("present", {None: " understand the", "3ps": " understands the", "present-participle": " understanding the", "past-participle": " understood the"}),
    "RPBD": ("past", {None: " understood", "root": " understand", "present-participle": " understanding", "past-participle": " understood"}),
    "RPBTD": ("past", {None: " understood the", "root": " understand the", "present-participle": " understanding the", "past-participle": " understood the"}),

    # Z - To use
    "Z": ("present", {None: " use", "3ps": " uses", "present-participle": " using", "past-participle": " used"}),
    "DZ": ("past", {None: " used", "root": " use", "present-participle": " using", "past-participle": " used"}),
    # TZ -- Special case
    "TZ": ("present", " used to"),
    "TDZ": ("past", " used to"),

    # P: To want (to)
    "P": ("present", {None: " want", "3ps": " wants", "present-participle": " wanting", "past-participle": " wanted"}),
    "PT": ("present", {None: " want to", "3ps": " wants to", "present-participle": " wanting to", "past-participle": " wanted to"}),
    "PD": ("past", {None: " wanted", "root": " want", "present-participle": " wanting", "past-participle": " wanted"}),
    "PTD": ("past", {None: " wanted to", "root": " want to", "present-participle": " wanting to", "past-participle": " wanted to"}),

    # RBGS - will - Auxiliary verb
    # These do not combine naturally with middle/structures.
    "RBGS": ("present", " will"),
    "RBGSZ": ("past", " would"),

    # RBS - To wish (to)
    "RBS": ("present", {None: " wish", "3ps": " wishes", "present-participle": " wishing", "past-participle": " wished"}),
    "RBTS": ("present", {None: " wish to", "3ps": " wishes to", "present-participle": " wishing to", "past-participle": " wished to"}),
    "RBSZ": ("past", {None: " wished", "root": " wish", "present-participle": " wishing", "past-participle": " wished"}),
    "RBTSDZ": ("past", {None: " wished to", "root": " wish to", "present-participle": " wishing to", "past-participle": " wished to"}),

    # RBG - To work (on)
    "RBG": ("present", {None: " work", "3ps": " works", "present-participle": " working", "past-participle": " worked"}),
    "RBGT": ("present", {None: " work on", "3ps": " works on", "present-participle": " working on", "past-participle": " worked on"}),
    "RBGD": ("past", {None: " worked", "root": " work", "present-participle": " working", "past-participle": " worked"}),
    "RBGTD": ("past", {None: " worked on", "root": " work on", "present-participle": " working on", "past-participle": " worked on"}),
}


def lookup(key):
    starter_lookup, middle_lookup, structure_lookup, ender_lookup = determine_parts(key[0])

    starter, verb_form, _ = starter_lookup
    tense, verb = ender_lookup
    middle_word, middle_verb_form = lookup_data(middle_lookup, tense, verb_form)
    structure_format, use_middle_verb_form, structure_verb_update = lookup_data(structure_lookup, tense, verb_form)

    original_verb_form = verb_form
    if use_middle_verb_form and middle_verb_form:
        verb_form = middle_verb_form

    middle_phrase = lookup_data(structure_format, tense, original_verb_form+"-"+verb_form).replace(
        '*', middle_word, 1).replace('!', starter)

    if structure_verb_update:
        verb_form = structure_verb_update

    ending = lookup_data(verb, verb_form)
    if ending == None:
        raise KeyError

    return middle_phrase + ending

def determine_parts(stroke):
    if stroke in NON_PHRASE_STROKES:
        raise KeyError

    match = PARTS_MATCHER.match(stroke)
    starter_key, v1, star, v2, f, ender_key = match.groups()
    if not match:
        raise KeyError

    ender_lookup = ENDERS.get(ender_key)
    if not ender_lookup:
        raise KeyError

    # Check short form first.
    simple_starter_lookup = SIMPLE_STARTERS.get(starter_key + v1)
    simple_pronoun_lookup = SIMPLE_PRONOUNS.get(star + v2)
    if simple_starter_lookup and simple_pronoun_lookup:
        simple_structure = SIMPLE_STRUCTURES[f]
        return simple_pronoun_lookup, simple_starter_lookup, simple_structure, ender_lookup

    # Full form lookup.
    starter_lookup = STARTERS.get(starter_key)
    if not starter_lookup:
        raise KeyError

    _, _, valid_enders = starter_lookup
    if valid_enders and ender_key not in valid_enders:
        raise KeyError

    middle_lookup = MIDDLES[v1 + star]

    structure_lookup = STRUCTURE_EXCEPTIONS.get(starter_key + v1 + star + v2 + f)
    if not structure_lookup:
        structure_lookup = STRUCTURE_EXCEPTIONS.get(v1 + star + v2 + f)
    if not structure_lookup:
        structure_lookup = STRUCTURES[star + v2 + f]

    return starter_lookup, middle_lookup, structure_lookup, ender_lookup


def lookup_data(data, *keys):
    for key in keys:
        data = _lookup_data(data, key)
    return data

def _lookup_data(data, key):
    if type(data) is not dict:
        return data

    result = data.get(key)
    if result != None:
        return result

    if '-' in key:
        key = key.split('-')[1]
        result = data.get(key)
        if result != None:
            return result

    if key[0] == 'b':
        result = data.get(key[1:])
        if result != None:
            return result

    return data.get(None)

# Proper reverse lookup support.
#
# This will show phrasing strokes in Plover's suggestions window.


REVERSE_STARTERS = {"": {"": True}}
REVERSE_MIDDLES = {}
REVERSE_STRUCTURES = {}
REVERSE_ENDERS = {}

for key in STARTERS:
    word = STARTERS[key][0]
    REVERSE_STARTERS.setdefault(word, {})
    REVERSE_STARTERS[word][key] = "full"

for key in SIMPLE_PRONOUNS:
    word = SIMPLE_PRONOUNS[key][0]
    REVERSE_STARTERS.setdefault(word, {})
    REVERSE_STARTERS[word][key] = "simple"

POSSIBLE_REVERSE_MATCH = re.compile(r"[a-zI ']+")
HYPHEN_OMIT_PATTERN = re.compile(r"[AO*EU-]")


def add_reverse_middles(stroke, data, form):
    if type(data) is dict:
        for k in data:
            add_reverse_middles(stroke, data[k], form)
        return

    word = data[0].strip()

    REVERSE_MIDDLES.setdefault(form, {})
    REVERSE_MIDDLES[form].setdefault(word, {})
    REVERSE_MIDDLES[form][word][stroke] = True


def add_reverse_structures(stroke, data, form):
    if type(data) is dict:
        for k in data:
            add_reverse_structures(stroke, data[k], form)
        return

    word = data.strip()

    REVERSE_STRUCTURES.setdefault(form, {})

    REVERSE_STRUCTURES[form].setdefault(word, {})
    REVERSE_STRUCTURES[form][word][stroke] = form

    # To lookup empty starters, have entries that don't include '!'
    word = word.replace('!', '').strip()
    REVERSE_STRUCTURES[form].setdefault(word, {})
    REVERSE_STRUCTURES[form][word][stroke] = form


def add_reverse_enders(stroke, data):
    if type(data) is dict:
        for k in data:
            add_reverse_enders(stroke, data[k])
        return

    word = data.strip()
    REVERSE_ENDERS.setdefault(word, {})
    REVERSE_ENDERS[word][stroke] = True


for key in MIDDLES:
    add_reverse_middles(key, MIDDLES[key], "full")

for key in SIMPLE_STARTERS:
    add_reverse_middles(key, SIMPLE_STARTERS[key], "simple")

for key in STRUCTURE_EXCEPTIONS:
    add_reverse_structures(key, STRUCTURE_EXCEPTIONS[key][0], "full")

for key in STRUCTURES:
    add_reverse_structures(key, STRUCTURES[key][0], "full")

for key in SIMPLE_STRUCTURES:
    add_reverse_structures(key, SIMPLE_STRUCTURES[key][0], "simple")

for key in ENDERS:
    add_reverse_enders(key, ENDERS[key][1])


def add_verb_stroke(prefix, suffix):
    # This check prevents adding in extra '-' when it will cause issues in lookup().
    if HYPHEN_OMIT_PATTERN.search(prefix) or HYPHEN_OMIT_PATTERN.search(suffix):
        return prefix + suffix

    return prefix + '-' + suffix


def reverse_match(result, full_text, stroke):
    try:
        if lookup([stroke]).strip() == full_text:
            result.append((stroke,))
    except KeyError:
        pass


def reverse_verb_match(result, full_text, text, prefix):
    if text not in REVERSE_ENDERS:
        return

    prefix = prefix.replace('**', '*', 1)
    for stroke in REVERSE_ENDERS[text]:
        reverse_match(result, full_text, add_verb_stroke(prefix, stroke))


def reverse_structure_match(result, full_text, text, prefix, form):
    words = text.split(' ')
    for i in range(len(words)+1):
        phrase = ' '.join(words[:i])
        if phrase not in REVERSE_STRUCTURES[form]:
            continue

        for structure_stroke in REVERSE_STRUCTURES[form][phrase]:
            remainder = text.replace(phrase, '', 1).strip()
            stroke = add_verb_stroke(prefix, structure_stroke)
            reverse_verb_match(result, full_text, remainder, stroke)


def reverse_middle_match(result, full_text, text, prefix, form):
    for word in REVERSE_MIDDLES[form]:
        if word in text:
            r = text.replace(word, '*', 1)
            r = r.replace(' *', '*')
            for s in REVERSE_MIDDLES[form][word]:
                stroke = add_verb_stroke(s, prefix) if form == "simple" else add_verb_stroke(prefix, s)
                reverse_structure_match(result, full_text, r, stroke, form)

    reverse_structure_match(result, full_text, text, prefix, form)


def reverse_lookup(text):
    if not POSSIBLE_REVERSE_MATCH.fullmatch(text):
        return []

    # Maximum phrase is 7 words, so early out if being asked for the stroke
    # for more.
    words = text.split(' ')
    if len(words) > 7:
        return []

    result = []

    for word in REVERSE_STARTERS:
        if word == "":
            continue

        if word in text:
            remainder = re.sub("\\b%s\\b" % word, '!', text, 1)
            strokes = REVERSE_STARTERS[word]
            for stroke in strokes:
                reverse_middle_match(result, text, remainder, stroke, strokes[stroke])

    # Testing for empty starters
    for stroke in REVERSE_STARTERS['']:
        reverse_middle_match(result, text, text, stroke, "full")

    return result




# Emily's Symbol Dictionary
import re

# define your starters here
uniqueStarters = ["SKWH"]

# define if attachment keys define where "space"s or "attachment"s lie
attachmentMethod = "space"

LONGEST_KEY = 1

# variant format = ['', 'E', 'U', 'EU']
# if no variants exist, then a single string can be used for the symbol and the variant specifier keys will be valid but ignored
symbols = {
    "SKWH": {
        # more computer function-y symbols
        "FG"    : ["{#Tab}", "{#Backspace}", "{#Delete}", "{#Escape}"],
        "RPBG"  : ["{#Up}", "{#Left}", "{#Right}", "{#Down}"],
        "FRPBG" : ["{#Page_Up}", "{#Home}", "{#End}", "{#Page_Down}"],
        "FRBG"  : ["{#AudioPlay}", "{#AudioPrev}", "{#AudioNext}", "{#AudioStop}"],
        "FRB"   : ["{#AudioMute}", "{#AudioLowerVolume}", "{#AudioRaiseVolume}", "{#Eject}"],
        ""      : ["", "{*!}", "{*?}", "{#Space}"],
        "FL"    : ["{*-|}", "{*<}", "{<}", "{*>}"],
        "RPB"   : ["↵", "↵", "↵", "↵"],
        # typable symbols
        "FR"     : ["!", "¬", "↦", "¡"],
        "FP"     : ["\"", "“", "”", "„"],
        "FRLG"   : ["#", "©", "®", "™"],
        "RPBL"   : ["$", "¥", "€", "£"],
        "FRPB"   : ["%", "‰", "‱", "φ"],
        "FBG"    : ["&", "∩", "∧", "∈"],
        "F"      : ["'", "‘", "’", "‚"],
        "FPL"    : ["(", "[", "<", "\{"],
        "RBG"    : [")", "]", ">", "\}"],
        "L"      : ["*", "∏", "§", "×"],
        "G"      : ["+", "∑", "¶", "±"],
        "B"      : [",", "∪", "∨", "∉"],
        "PL"     : ["-", "−", "–", "—"],
        "R"      : [".", "•", "·", "…"],
        "RP"     : ["/", "⇒", "⇔", "÷"],
        "LG"     : [":", "∋", "∵", "∴"],
        "RB"     : [";", "∀", "∃", "∄"],
        "PBLG"   : ["=", "≡", "≈", "≠"],
        "FPB"    : ["?", "¿", "∝", "‽"],
        "FRPBLG" : ["@", "⊕", "⊗", "∅"],
        "FB"     : ["\\", "Δ", "√", "∞"],
        "RPG"    : ["^", "«", "»", "°"],
        "BG"     : ["_", "≤", "≥", "µ"],
        "P"      : ["`", "⊂", "⊃", "π"],
        "PB"     : ["|", "⊤", "⊥", "¦"],
        "FPBG"   : ["~", "〜", "⊇", "˜"],
        "FPBL"   : ["↑", "←", "→", "↓"]
    },
    "#SKWH": {
        # add your own strokes here (or above, or wherever else you like)!
        ""       : "test"
    }
}


def lookup(chord):

    # normalise stroke from embedded number, to preceding hash format
    stroke = chord[0]
    if any(k in stroke for k in "1234506789"):  # if chord contains a number
        stroke = list(stroke)
        numbers = ["O", "S", "T", "P", "H", "A", "F", "P", "L", "T"]
        for key in range(len(stroke)):
            if stroke[key].isnumeric():
                stroke[key] = numbers[int(stroke[key])]  # set number key to letter
                numberFlag = True
        stroke = ["#"] + stroke
        stroke = "".join(stroke)

    # the regex decomposes a stroke into the following groups/variables:
    # starter                #STKPWHR
    # attachments                         AO
    # capitalisation                             */-
    # variants                                          EU
    # pattern                                                  FRPBLG
    # repetitions                                                         TS
    #                                       (unused: DZ)
    match = re.fullmatch(r'([#STKPWHR]*)([AO]*)([*-]?)([EU]*)([FRPBLG]*)([TS]*)', stroke)

    if match is None:
        raise KeyError
    (starter, attachments, capitalisation, variants, pattern, repetitions) = match.groups()

    if starter not in uniqueStarters:
        raise KeyError
    if len(chord) != 1:
        raise KeyError
    assert len(chord) <= LONGEST_KEY

    # calculate the attachment method, and remove attachment specifier keys
    attach = [(attachmentMethod == "space") ^ ('A' in attachments),
              (attachmentMethod == "space") ^ ('O' in attachments)]

    # detect if capitalisation is required, and remove specifier key
    capital = capitalisation == '*'

    # calculate the variant number, and remove variant specifier keys
    variant = 0
    if 'E' in variants:
        variant = variant + 1
    if 'U' in variants:
        variant = variant + 2

    # calculate the repetition, and remove repetition specifier keys
    repeat = 1
    if 'S' in repetitions:
        repeat = repeat + 1
    if 'T' in repetitions:
        repeat = repeat + 2

    if pattern not in symbols[starter]:
        raise KeyError

    # extract symbol entry from the 'symbols' dictionary, with variant specification if available
    selection = symbols[starter][pattern]
    if type(selection) == list:
        selection = selection[variant]
    output = selection

    # repeat the symbol the specified number of times
    output = output * repeat

    # attachment space to either end of the symbol string to avoid escapement,
    # but prevent doing this for retrospective add/delete spaces, since it'll
    # mess with these macros
    if selection not in ["{*!}", "{*?}"]:
        output = " " + output + " "

    # add appropriate attachment as specified (again, prevent doing this
    # for retrospective add/delete spaces)
    if selection not in ["{*!}", "{*?}"]:
        if attach[0]:
            output = "{^}" + output
        if attach[1]:
            output = output + "{^}"

    # cancel out some formatting when using space attachment
    if attachmentMethod == "space":
        if not attach[0]:
            output = "{}" + output
        if not attach[1]:
            output = output + "{}"

    # apply capitalisation
    if capital:
        output = output + "{-|}"

    # all done :D
    return output
